#' prepare SWIM gages
#' @description reads SWIM gageinfo dbf as created in gage info pre process and
#' swim points data to create a set of SWIM gages that will be used later.
#' @param SWIM_dbf_file character path tp summary file as created in gageinfo step.
#' @param SWIM_points_file character path to SWIM source data
#' @param ref_gages sf data.frame gages from USGS nldi dataset / ref gages
#' @param gages_add character vector of gages to not filter out. 
#' @param POR_start Date to use as period of record start date
#' @param min_obs minimum number of observations 
#' @return list of three data frames for use later and in mapping output of this 
#' function.
prepare_SWIM <- function(SWIM_dbf_file, SWIM_points_file, ref_gages,
                         gages_add,
                         POR_start = as.Date("1979-10-01"), 
                         min_da_km_gages = 5,
                         min_obs = 90) {
  # Read gagesIII dbf that is a POR summary produced from 00_gage_info.Rmd
  gagesIII_db_full <- read.dbf(file = SWIM_dbf_file, as.is = TRUE) %>%
    rename(total_obs = count_nu, total_poss_obs = TPdays1)
  
  # Read in shapefile
  gagesIII_sf_full <- read_sf(SWIM_points_file, "SWIM_gage_loc") %>%
    st_zm(drop = TRUE) %>%
    # Use site_no as the authoritative station ID name
    rename(site_no = Gage_no) %>%
    # Drop fields in NLDI gages
    select(-c(COMID, REACHCODE, REACH_meas)) %>%
    # Join to NLDI reference gages
    left_join(select(st_drop_geometry(ref_gages), id, comid = nhdpv2_COMID, 
                     reachcode = nhdpv2_REACHCODE, 
                     reach_meas = nhdpv2_REACH_measure), 
              by = c("site_no" = "id")) 
  
  # Read Gages III shapefile
  gagesIII_sf <- gagesIII_sf_full %>% 
    # Get rid of invalid COMIDs, for multiple site IDs get one with the furthest reach measure downstream
    filter(comid > 0, site_no %in% gagesIII_db_full$site_no) %>% 
    # There are some rows with dupslicate site_no, drainage area, and POR infomrmation
    group_by(site_no) %>% 
    slice(which.min(reach_meas)) %>%
    # There were some duplicate rows that needed to be removed
    select(site_no, comid, reachcode, reach_meas) %>% 
    distinct()
  
  # Start documenting gages that are dropped out; these gages have no mean daily Q
  document_dropped_gage(filter(select(st_drop_geometry(gagesIII_sf_full), site_no), 
                               !site_no %in% gagesIII_sf$site_no),
                        "GAGES-III", "No Q or COMID")
  
  # First subset
  gagesIII_db <- gagesIII_db_full %>%
    # Get rid of inital fields we are not interested in, 
    select(-c(agency_cd, ts_id)) %>% 
    #     Constrain by date and/or POR 
    filter(end_date > POR_Start | total_obs > min_obs, 
           #     remove site codes associated with canals, ditchest, etc
           !site_tp_cd %in% c("ST-CA", "ST-DCH", "FA-STS"), 
           #     keep sites with daily mean streamflow parameter
           stat_cd == "00003") %>% 
    distinct() 
  
  # Gages that don't fit the POR requirement, or are non-ST site types
  document_dropped_gage(filter(select(st_drop_geometry(gagesIII_sf), site_no), 
                               !site_no %in% gagesIII_db$site_no),
                        "GAGES-III", "Non-ST; POR")
  
  # scan for potential canals/ditches among the remaining ST site types
  potCanalsIII <- gagesIII_db %>%
    # Flag station_nm with the follwing words
    filter(grepl(paste(c("DITCH", "ditch", "CANAL", "canal", "CN", "DIV", "div"), 
                       collapse = "|"), station_nm)) %>%
    # Remove the station_nm with the following
    filter(!grepl(paste(c("RIVER AT", "RIVER BELOW", "RIVER BL", "R AB", "R BL", "R ABV"), 
                        collapse = "|"), 
                  station_nm, ignore.case = TRUE)) %>%
    # These have been manually checked and are located in gage_mod.R
    filter(!site_no %in% gages_add) 
  
  # Write out for manual checking
  if(!file.exists("temp/GagestoCheck_GAGESIII.csv")){
    write.csv(potCanalsIII, "temp/GagestoCheck_GAGESIII.csv")
  }
  
  # Subset further by streamgages below the minimum drainage area critera
  gagesIII_db_fin <- gagesIII_db %>% 
    filter(!site_no %in% potCanalsIII$site_no, drain_area > (min_da_km_gages * sqkm_sqmi))
  
  # Write out gages that dropped to document.
  document_dropped_gage(select(potCanalsIII, site_no), "GAGES-III", "ST; Pot. Canal")
  document_dropped_gage(gagesIII_db %>%
                          filter(!site_no %in% potCanalsIII$site_no, 
                                 drain_area < min_da_km_gages) %>% 
                          select(site_no),
                        "GAGES-III", "below DA")
  
  # Join shapefile to POR table
  gagesIII_final <- gagesIII_sf %>% 
    inner_join(gagesIII_db_fin, by = "site_no") %>%
    select(-c(coord_acy_, huc_cd, parm_cd, stat_cd, missing_da, site_tp_cd)) %>%
    mutate(source = "gagesIII") %>%
    # deal with one duplicate row
    group_by(site_no) %>%
    filter(total_obs == max(total_obs)) %>%
    ungroup() %>%
    st_transform()
  
  return(list(final = gagesIII_final, full = gagesIII_sf_full, filtered = gagesIII_sf))
}

prepare_gage_location <- function(NHDPlusV2_gageloc_file, NHDPlusV2_gdb,
                                  SWIM, ref_gages,
                                  min_da_km_gages = 5) {
  # Read gages3 dbf that is a POR summary produced from 00_gage_info.Rmd
  gageloc_db_full <- read.dbf(file = NHDPlusV2_gageloc_file, as.is = TRUE) %>%
    rename(total_obs = count_nu, total_poss_obs = TPdays1)
  
  # Read in shapefile
  gageloc_sf_full <- read_sf(NHDPlusV2_gdb, "Gage") %>%
    # Drop Z geometries
    st_zm(drop = TRUE) %>%
    rename(site_no = SOURCE_FEA) %>%
    select(-c(FLComID, REACHCODE, Measure)) %>%
    # Disregard locations already removed in GAGESIII
    filter(!site_no %in% SWIM$full$site_no) %>%
    # Update index information with NLDI gages
    left_join(select(st_drop_geometry(ref_gages), site_no = id, 
                     comid = nhdpv2_COMID, reachcode = nhdpv2_REACHCODE, 
                     reach_meas = nhdpv2_REACH_measure ), by = "site_no")
  
  # Subset to locations with streamflow
  gageloc_sf <- gageloc_sf_full %>% 
    # Get rid of invalid COMIDs, for multiple site IDs get one with the furthest reach measure downstream
    filter(site_no %in% gageloc_db_full$site_no) %>% 
    # Like GAGES-III, multiple site_no 
    group_by(site_no) %>% 
    slice(which.min(reach_meas)) %>%
    # There were some duplicate rows that needed to be removed
    select(site_no, comid, reachcode, reach_meas) %>% 
    distinct()
  
  # Start documenting gages that are dropped out; these gages have no mean daily Q
  document_dropped_gage(filter(select(st_drop_geometry(gageloc_sf_full), site_no), 
                               !site_no %in% gageloc_sf$site_no), 
                        "GageLoc", "No Q, COMID")
  
  # do the same exercise with GageLoc, but with more site type attributes to filter out
  gageloc_db <- gageloc_db_full %>%
    filter(end_date > POR_Start | total_obs > min_obs, 
           !site_tp_cd %in% 
             c("ST-CA", "ST-DCH", "FA-STS", "SB_TSM", "FA-DV", "FA-STS", "FA-CS",
               "FA-WWD", "FA-OF", "GW"), stat_cd == "00003") %>%
    # Some sites are listed twice with same information but different agencies, 
    select(-c(agency_cd, ts_id))  %>% 
    distinct(site_no, drain_area, dec_lat_va, dec_long_v, .keep_all = T)
  
  # Gages that don't fit the POR requirement, or are non-ST site types
  document_dropped_gage(filter(select(gageloc_db_full, site_no), 
                               !site_no %in% gageloc_db$site_no), 
                        "GageLoc", "Non-ST; POR")
  
  # scan for potential canals/ditches 
  potCanals_gageloc <- gageloc_db %>% 
    filter(grepl(paste(c("DITCH", "ditch", "CANAL", "canal", "CN", "DIV", "div"), 
                       collapse = "|"), station_nm)) %>%
    filter(!site_no %in% gages_add)
  
  # Write out for manual checking
  if(!file.exists("temp/GagestoCheck_GageLoc.csv")){
    write.csv(potCanals_gageloc, "temp/GagestoCheck_GageLoc.csv")
  } else {
    potCanals_gageloc <- read.csv("temp/GagestoCheck_GageLoc.csv")
  }
  
  gageloc_db_fin <- gageloc_db %>% 
    filter(!site_no %in% potCanals_gageloc$site_no, drain_area > 
             (min_da_km_gages * sqkm_sqmi)) 
  
  # Gages that are ST site types, but reside on canals, artificial waterbodies.
  document_dropped_gage(select(potCanals_gageloc, site_no), "GageLoc", "ST; Pot. Canal")
  document_dropped_gage(gageloc_db_fin %>%
                          filter(!site_no %in% potCanals_gageloc$site_no, 
                                 drain_area < min_da_km_gages) %>% select(site_no),
                        "GageLoc", "below DA")
  
  # NHD gages shpafile
  gageloc_final <-  gageloc_sf %>% 
    inner_join(gageloc_db_fin, by = "site_no") %>% 
    select(-c(coord_acy_, huc_cd, parm_cd, stat_cd, missing_da, site_tp_cd)) %>% 
    st_as_sf(coords = c("dec_long_v", "dec_lat_va")) %>%
    mutate(source = "GageLoc") %>%
    # deal with one duplicate row
    group_by(site_no) %>%
    filter(total_obs == max(total_obs)) %>%
    ungroup() %>%
    st_sf() %>%
    st_compatibalize(SWIM$final)
}